#!/usr/bin/env python

"""
Module to support any and all Extract Transform Load operations from
SiriusXm into PySoSirius Historical DB. 
"""

# System packages
from copy import deepcopy
import logging

# installed packages
import pymongo
from pysosirius.sirius_playing import SiriusCurrentlyPlaying
from pysosirius.common.exceptions import AttributeNotFoundError 

# global, module level logger
log = logging.getLogger(__name__)

class ETL(object):
	"""
	Class for encapsulating all thing ETL for a channel

	Attributes:
        pss_channel (SiriusChannel): Class for channel metadata.
        database (MongoClient): DB connection.
        database_write (bool): True to write to DB.
        collection_name (str): name of channel collection in DB.
        __last_played (:obj:`Dict`, optional): Dict of last played song.

		__init__

        Initialize a wrapper for one channel's ETL functions

        Args:
            pss_channel (SiriusChannel): Sirius channel metadata
            database (:obj:`MongoClient`, optional): DB connex
            database_write (bool): write to db


	"""
	def __init__(self,
				pss_channel,
				database = None,
				database_write = True):
		super(ETL, self).__init__()

		# args
		self.pss_channel = pss_channel
		self.database = database
		self.database_write = database_write

		# computed values
		self.collection_name = 'all_songs'

		# internal data
		self.__last_played = None

		# internal setup
		self.__setup_db_structures()
		self.__update_last_played()

	@property
	def last_played_start(self):
		"""str: datetime of last played song"""

		if self.__last_played:
			# get the last datetime logged for the channel of interest
			_,channel_dict = self.__get_channel_dict(self.__last_played,self.pss_channel.channel)
			if channel_dict:
				return channel_dict['datetimes'][-1]

		return None

	def __get_channel_dict(self,song_doc,channel_number):
		""" return the offset and the dict itself else (none,none) """

		# get the channel dictionary of interest
		try:
			return [(i,channel_dict) for i,channel_dict in enumerate(song_doc['channels']) if channel_dict['number'] == channel_number][0]
		except:
			return None,None

	def __update_last_played(self,last_played = {}):
		"""
		Update the last played data member

		Args:
		    last_played: Optional last played document

		Returns:
		    None
		"""

		if last_played:
			self.__last_played = last_played
		elif self.database:

			if 'last_played' in self.database.collection_names():

				self.__last_played = self.database['last_played'].find_one(
											{'_id':self.pss_channel.channel})
	
	def __setup_db_structures(self):
		"""
		Set up db structures for the channel's collection
		"""

		# create channel based indexes
		if self.database and self.database_write:

			collection = self.database[self.collection_name]

			collection.create_index([('artist.name',pymongo.TEXT),
									 ('song.name',pymongo.TEXT),
									 ('channels.number',pymongo.TEXT),
									 ('channels.name',pymongo.TEXT),
									 ('channels.datetimes', pymongo.DESCENDING)])
	def get_existing(self):
		"""
		Get an existing document int he DB

		Args:
		    None

		Returns:
		    Dict of existing document
		"""

		existing_doc = None

		if self.database:

			# criteria to search index for song match
			existing_filter = {'artist.name': self.pss_channel.currently_playing.artist_name,
								'song.name': self.pss_channel.currently_playing.song_name
								}

			# search english, case insensitive (1 is less strict)
			# TODO: what about spanish channels?
			existing_collation = {'locale': "en",
									'strength': 2
									}

			try:
				# will raise IndexError if no item in the collection exists
				existing_doc = self.database[self.collection_name].find(
								filter=existing_filter).collation(
									existing_collation)[0]
			except IndexError:
				pass
		
		return existing_doc

	def extract_transform_load(self):
		"""
		Driver for the ETL process
		"""

		# if the extraction was successful
		if self.__extract():
			new_doc, doc_updates = self.__transform()

			# theres always a new doc for newly switched to song
			if new_doc:
				log.info('\nArtist: %s \nSong: %s',
					self.pss_channel.currently_playing.artist_name,
					self.pss_channel.currently_playing.song_name)
				self.__load(new_doc,doc_updates)
		else:
			log.info('Extracting Data for Channel: %s \n %s \n %s \n',
				self.pss_channel.name,
				self.pss_channel.url,
				self.pss_channel.currently_playing._get_url())

	def __extract(self):
		"""
		Hit sirius server for currently playing
		"""
		try:
			# will raise AttributeNotFoundError
			self.pss_channel.get_currently_playing()
			# return status of extraction
			self.pss_channel.currently_playing.status
		except AttributeNotFoundError:
			return False

		# return status of extraction
		return self.pss_channel.currently_playing.status

	def __transform(self):
		"""
		Merge XM data and local document structure
		"""

		# helper function to manipulate documents
		def transform_helper(extraction_data,existing_transformed_data=None):
			"""
			Transform an XM doc structure to the new structure or
			merge an XM doc structure and an existing together

			Args:
			    extraction_data: XM data struct
			    existing_transformed_data: to be stored, existing struct

			Returns:
			    Dict of new document
			"""

			ROOT_SONG_KEYS = SiriusCurrentlyPlaying.JSON_ROOT_MSG_DATA_KEYS['song']
			ATTRIBUTE_KEYS = SiriusCurrentlyPlaying.JSON_MSG_DATA_KEYS
			SONG_START_TIME = ROOT_SONG_KEYS + ATTRIBUTE_KEYS['start']

			if existing_transformed_data:

				new_transformed_data = deepcopy(existing_transformed_data)
			else:
				new_transformed_data = {}


			# index based fields
			new_transformed_data['album'] = {'name': SiriusCurrentlyPlaying.get_attribute(ROOT_SONG_KEYS+ATTRIBUTE_KEYS['album'],extraction_data)}
			new_transformed_data['song'] = {'name':  SiriusCurrentlyPlaying.get_attribute(ROOT_SONG_KEYS+ATTRIBUTE_KEYS['song'],extraction_data)}
			new_transformed_data['artist'] = {'name':  SiriusCurrentlyPlaying.get_attribute(ROOT_SONG_KEYS+ATTRIBUTE_KEYS['artist'],extraction_data)}
			
			# update list of channels
			channel_data = {'id': SiriusCurrentlyPlaying.get_attribute(['channelMetadataResponse','metaData','channelId'],extraction_data),
							'number': SiriusCurrentlyPlaying.get_attribute(['channelMetadataResponse','metaData','channelNumber'],extraction_data),
							'name': SiriusCurrentlyPlaying.get_attribute(['channelMetadataResponse','metaData','channelName'],extraction_data)}

			# get the new played datetime
			new_datetime = SiriusCurrentlyPlaying.get_attribute(SONG_START_TIME,extraction_data)

			# get the channel dictionary
			channel_dict_offset,existing_channel_dict = self.__get_channel_dict(new_transformed_data,channel_data['number'])

			if existing_channel_dict:
				# add the new date time to the end
				existing_channel_dict['datetimes'] = existing_channel_dict['datetimes'] + [new_datetime]
				# update the existing offset with the new data
				new_transformed_data['channels'][channel_dict_offset] = existing_channel_dict
			else:
				# add the first occurence
				channel_data['datetimes'] = [new_datetime]

				#print(channel_data['datetimes'])
				#print("*"*10)
				#print(new_transformed_data.get('channels',[]))
				# add the new data blindly, its okay since its new
				new_transformed_data['channels'] = new_transformed_data.get('channels',[]) + [channel_data]	

			image_list = []
			for artwork in SiriusCurrentlyPlaying.get_attribute(['channelMetadataResponse','metaData','currentEvent','song','creativeArts'],extraction_data):

				image_data = {}

				if (artwork.get('type',None) == 'IMAGE'
					and artwork.get('encrypted',None) == False):
					
					image_data['size'] = artwork['size']
					image_data['url'] = artwork['url']
					image_list.append(image_data)
					
			new_transformed_data['artwork'] = image_list

			return new_transformed_data

		# begin the transform
		if self.last_played_start != self.pss_channel.currently_playing.start:
			
			existing_doc = self.get_existing()

			if existing_doc:

				# generate the updated doc
				new_doc = transform_helper(self.pss_channel.currently_playing.data,existing_doc)

				doc_updates = {}
				#doc_updates.clear()

				# index values that should never change once a doc has been created
				# or we don't care to write their updates
				for key,value in new_doc.items():
					if key not in ['_id','album','song','artist','artwork']:
						doc_updates[key] = value

				return (new_doc,doc_updates)

			else:
				# we have a new song so insert it into the collection
				new_doc = transform_helper(self.pss_channel.currently_playing.data)
				return (new_doc,None)

		# its the same song, no new doc/no updates
		return (None,None)

	def __load(self,new_doc,doc_updates = {}):
		"""
		Save the new document or updated document to the db

		Args:
		    new_doc: newly formatted doc to be stored
		    doc_updates: dic of keywords to update an existing doc

		Returns:
		    None
		"""

		# make sure we log the last played on the channel
		# regardless of writing to the db or not
		self.__update_last_played(new_doc)

		#from pprint import pprint

		#pprint(new_doc)
		#pprint(doc_updates)

		if self.database and self.database_write:

			# update the channel specific song documents
			if doc_updates:

				# criteria to search index for song match
				existing_filter = {'artist.name': self.pss_channel.currently_playing.artist_name,
									'song.name': self.pss_channel.currently_playing.song_name
									}

				# search english, case insensitive (1 is less strict)
				# TODO: what about spanish channels?
				existing_collation = {'locale': "en",
										'strength': 2
										}

				self.database[self.collection_name].update_one(existing_filter,
																{'$set': doc_updates},
																collation=existing_collation)
			else:
				self.database[self.collection_name].insert_one(new_doc)

			# update the last played collection doc for this channel
			new_doc['_id'] = self.pss_channel.channel
			last_played_filter = {'_id':self.pss_channel.channel}

			# replace the old last song with the new,last song
			self.database[u'last_played'].replace_one(last_played_filter,
														new_doc,upsert=True)



#!/usr/bin/env python

""" 
Used in conjunction with PySoSirius package (necessary channel metadata).
Useful for compiling currently playing song data from Sirius XM 
and storing it a document DB for historical records to be queried later. 

Poll for all PySoSirius channels (sqlite db) with no logging
	python poll_channels.py

Poll specific channels with no logging
	python poll_channels.py -C 56 53

Poll channels with logging
	python poll_channels.py -C 56 -ll INFO

Poll channels no db interaction - useful for testing
for just to getting currently playing data with no need or use
for more than last song played and current song playing
	python poll_channels.py -C 56 -ll DEBUG --db_write_disabled

"""
import datetime
import time
import threading
import logging
from argparse import ArgumentParser

import pymongo

from pysosirius.pysosirius import PySoSirius
from etl.extract_transform_load import ETL

# global, module level logger
log = logging.getLogger(__name__)

def extraction(etl_channel):
	"""
	Launch a thread to start compiling played songs per channel

	Args:
	    etl_channel: ETL - wrapper for etl of one channel

	Returns:
	    None

	Raises:
	    None
	"""
	while True:
		try:
			etl_channel.extract_transform_load()
		except:
			channel = etl_channel.pss_channel.channel
			url = etl_channel.pss_channel.currently_playing._get_url()
			error_string = '\n {0} {1} {0} \n{2}'.format('*'*25,channel,url)
			log.exception(error_string)
		time.sleep(60)

def get_pss_channels(channels=None):
	"""
	Given (optional) channel numbers, get the PySoSirius metadata

	Args:
	    channels: int - List of channel numbers.

	Returns:
	    list of SeriousChannel objects

	Raises:
	    None
	"""

	py_so_sirius = PySoSirius()
	pss_channels = []

	if channels:

		# we want specific channels
		for channel in channels:
			pss_channels.append(py_so_sirius.get_channel(channel=channel))
	else:

		# we want them all
		pss_channels = py_so_sirius.get_channels()

	return pss_channels

def start_channel_compile(channels,database=None,db_write_enabled=True):
	"""
	Launch a threads to start compiling played songs per channel

	Args:
	    channels: SiriusChannel - List of channel objects.
	    database: MongoClient - connection to specific db.
		db_write_enabled: bool - should writing be enabled

	Returns:
	    None

	Raises:
	    None
	"""

	for pss_channel in channels:

		etl_channel = ETL(pss_channel,
						database = database,
						database_write = db_write_enabled)

		# kick off the compile thread
		etl_thread = threading.Thread(name = pss_channel.channel,
									target = extraction,
									args = (etl_channel,))
		etl_thread.daemon = True
		etl_thread.start()

def main(**kwargs):
	"""
	Driver of command line interface. Sets up DB connx and channel metadata
	to prep it for compile.

	kwargs:
	    db_ip: str - ip address for the db connection 
	    db_port: int - port to connect to db
	    db_write_disabled: bool - should we be writing to the db

	Returns:
	    None

	Raises:
	    None
	"""

	# Connect to datbase server
	client = pymongo.MongoClient(kwargs['db_ip'],kwargs['db_port'])

	# connect to database
	database = client['channel_data']

	# get channels of interest
	pss_channels = get_pss_channels(kwargs['channels'])

	# Temporarily delete so we start from scratch
	#database['last_played'].delete_many({})
	#database['all_songs'].delete_many({})
	#database['all_songs'].drop_indexes()

	# start compile threads
	start_channel_compile(pss_channels,database,not(kwargs['db_write_disabled']))

	# loop infinitum - used to keep daemon threads running
	while True:
		log.info('Minute: %d',(1440 - datetime.datetime.now().minute))
		time.sleep(60)
		pass

if __name__ == '__main__':
	"""
	Main setup of capturing commandline args, and general setup of logging.
	"""

	parser = ArgumentParser(description='Populate Database from Siriux XM')

	# General args
	parser.add_argument('-ll', '--log_level',
						type=str,
						choices=['DEBUG','INFO','WARNING','ERROR','CRITICAL'],
						help='Set the logging level')

	parser.add_argument('-log', '--log_file',
						type=str,
						help='Set the error file for output')

	# Compile args
	parser.add_argument('-C', '--channels',
						type=int,
						nargs = '*',
						help='Override channels to compile')

	# DB Args
	parser.add_argument('-DBIP', '--db_ip',
						type=str,
						help='IP address for db connection',
						default = '127.0.0.1')

	parser.add_argument('-DBP', '--db_port',
						type=int,
						help='Port address for db connection',
						default = '27017')

	parser.add_argument('-DBW', '--db_write_disabled',
						action = 'store_true',
	 					help='True to disable writing to the db')

	# Get all args from commandline
	args = parser.parse_args()
 	args_dict = vars(args)

 	# Set up global logger
 	if args.log_level:
 		logging.basicConfig(filename=args.log_file)
 		logger = logging.getLogger()
		handler = logging.StreamHandler()
		formatter = logging.Formatter(
		        		'[%(levelname)s] [%(threadName)s] %(message)s')
		handler.setFormatter(formatter)
		logger.addHandler(handler)
		logger.setLevel(args.log_level)

 	# remove any args not needed
 	args_dict.pop('log_level')
 	args_dict.pop('log_file')

 	# pass remaining args to main so it can do what it needs
	main(**args_dict)


from setuptools import setup, find_packages

setup(
	name='pysosiriusmongo',
	version='0.1.1',
	description='PySoSirius SiriusXM Song Storage',
	url='https://github.com/T-Santos/PySoSiriusMongo',
	author='Tyler Santos',
	author_email='1tsantos7+pysosirius@Gmail.com',
	license='MIT',
	packages=['pysosiriusmongo','pysosiriusmongo.etl'],
	# Project uses BeautifulSoup, requests and selenium to
	# parse webpages for new info not stored in the local db
	install_requires=['pysosirius','pymongo'],
	keywords="Sirius SiriusXM XM Music MongoDB",	
	classifiers=[
		# How mature is this project? Common values are
		'Development Status :: 1 - Planning',

		# Indicate who your project is intended for
		'Intended Audience :: Developers',

		# Pick your license as you wish (should match "license" above)
		 'License :: OSI Approved :: MIT License',

		# Specify the Python versions you support here. In particular, ensure
		# that you indicate whether you support Python 2, Python 3 or both.
		'Programming Language :: Python :: 2.7'
	])
